@@ -14,3 +14,8 @@ else
 fi
 
 resolve_resources config/ $output_file noignore $image_prefix $tag
+
+# Update value of "devel" in labels to $tag
+if [[ -n ${tag} ]]; then
+    sed -i -r "s/\"?devel\"?$/${tag}/g" $output_file
+fi
