@@ -27,25 +27,28 @@ import (
 	"github.com/google/go-cmp/cmp/cmpopts"
 	"github.com/tektoncd/triggers/pkg/apis/triggers/v1alpha1"
 	listers "github.com/tektoncd/triggers/pkg/client/listers/triggers/v1alpha1"
-	"github.com/tektoncd/triggers/pkg/reconciler"
+	"go.uber.org/zap"
 	"golang.org/x/xerrors"
 	appsv1 "k8s.io/api/apps/v1"
 	corev1 "k8s.io/api/core/v1"
-	"k8s.io/apimachinery/pkg/api/equality"
 	"k8s.io/apimachinery/pkg/api/errors"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/labels"
 	"k8s.io/apimachinery/pkg/util/intstr"
-	"k8s.io/client-go/tools/cache"
+	"k8s.io/client-go/kubernetes"
+	appsv1lister "k8s.io/client-go/listers/apps/v1"
+	corev1lister "k8s.io/client-go/listers/core/v1"
+	"knative.dev/pkg/logging"
+	pkgreconciler "knative.dev/pkg/reconciler"
 
-	"knative.dev/pkg/controller"
+	triggersclientset "github.com/tektoncd/triggers/pkg/client/clientset/versioned"
+	eventlistenerreconciler "github.com/tektoncd/triggers/pkg/client/injection/reconciler/triggers/v1alpha1/eventlistener"
+	"knative.dev/pkg/ptr"
 )
 
 const (
-	// eventListenerAgentName defines logging agent name for EventListener Controller
-	eventListenerAgentName = "eventlistener-controller"
-	// eventListenerControllerName defines name for EventListener Controller
-	eventListenerControllerName = "EventListener"
+	// ControllerName defines the name for EventListener Controller
+	ControllerName = "EventListener"
 	// eventListenerConfigMapName is for the automatically created ConfigMap
 	eventListenerConfigMapName = "config-logging-triggers"
 	// eventListenerServicePortName defines service port name for EventListener Service
@@ -64,12 +67,12 @@ var (
 	// ElPort defines the port for the EventListener to listen on
 	ElPort = flag.Int("el-port", 8080,
 		"The container port for the EventListener to listen on.")
-	// PeriodSeconds defines Period Seconds for the EventListener Liveness Probe
+	// PeriodSeconds defines Period Seconds for the EventListener Liveness and Readiness Probes
 	PeriodSeconds = flag.Int("period-seconds", 10,
-		"The Period Seconds for the EventListener Liveness Probe.")
-	// FailureThreshold defines the Failure Threshold for the EventListener Liveness Probe
-	FailureThreshold = flag.Int("failure-threshold", 1,
-		"The Failure Threshold for the EventListener Liveness Probe.")
+		"The Period Seconds for the EventListener Liveness and Readiness Probes.")
+	// FailureThreshold defines the Failure Threshold for the EventListener Liveness and Readiness Probes
+	FailureThreshold = flag.Int("failure-threshold", 3,
+		"The Failure Threshold for the EventListener Liveness and Readiness Probes.")
 	// StaticResourceLabels is a map with all the labels that should be on
 	// all resources generated by the EventListener
 	StaticResourceLabels = map[string]string{
@@ -80,120 +83,118 @@ var (
 
 // Reconciler implements controller.Reconciler for Configuration resources.
 type Reconciler struct {
-	*reconciler.Base
+
+	// KubeClientSet allows us to talk to the k8s for core APIs
+	KubeClientSet kubernetes.Interface
+
+	// TriggersClientSet allows us to configure triggers objects
+	TriggersClientSet triggersclientset.Interface
+
 	// listers index properties about resources
+	configmapLister     corev1lister.ConfigMapLister
+	deploymentLister    appsv1lister.DeploymentLister
 	eventListenerLister listers.EventListenerLister
+	serviceLister       corev1lister.ServiceLister
+
+	// systemNamespace is the namespace where the reconciler is deployed
+	systemNamespace string
 }
 
-// Check that our Reconciler implements controller.Reconciler
-var _ controller.Reconciler = (*Reconciler)(nil)
+var (
+	// Check that our Reconciler implements eventlistenerreconciler.Interface
+	_ eventlistenerreconciler.Interface = (*Reconciler)(nil)
+	// Check that our Reconciler implements eventlistenerreconciler.Finalizer
+	_ eventlistenerreconciler.Finalizer = (*Reconciler)(nil)
+)
 
-// Reconcile compares the actual state with the desired, and attempts to
+// ReconcileKind compares the actual state with the desired, and attempts to
 // converge the two.
-func (c *Reconciler) Reconcile(ctx context.Context, key string) (returnError error) {
-	c.Logger.Infof("event-listener-reconcile %s", key)
-	// Convert the namespace/name string into a distinct namespace and name
-	namespace, name, err := cache.SplitMetaNamespaceKey(key)
-	if err != nil {
-		c.Logger.Errorf("invalid resource key: %s", key)
-		return nil
-	}
-
-	// Get the EventListener resource
-	original, err := c.eventListenerLister.EventListeners(namespace).Get(name)
-	if errors.IsNotFound(err) {
-		// The resource no longer exists, in which case we stop processing.
-		c.Logger.Infof("EventListener %q in work queue no longer exists", key)
-		cfgs, err := c.eventListenerLister.EventListeners(namespace).List(labels.Everything())
-		if err != nil {
-			return err
-		}
-		if len(cfgs) > 0 {
-			return nil
-		}
-		err = c.KubeClientSet.CoreV1().ConfigMaps(namespace).Delete(eventListenerConfigMapName, &metav1.DeleteOptions{})
-		if errors.IsNotFound(err) {
-			return nil
-		}
-		return err
-	} else if err != nil {
-		c.Logger.Errorf("Error retrieving EventListener %q: %s", name, err)
-		return err
-	}
-
-	// Don't modify the informer's copy
-	el := original.DeepCopy()
+func (r *Reconciler) ReconcileKind(ctx context.Context, el *v1alpha1.EventListener) pkgreconciler.Event {
 	// Initial reconciliation
-	if equality.Semantic.DeepEqual(el.Status, v1alpha1.EventListenerStatus{}) {
-		el.Status.InitializeConditions()
-		el.Status.Configuration.GeneratedResourceName = fmt.Sprintf("%s-%s", GeneratedResourcePrefix, el.Name)
-	}
+	el.Status.InitializeConditions()
+	el.Status.Configuration.GeneratedResourceName = fmt.Sprintf("%s-%s", GeneratedResourcePrefix, el.Name)
 
-	// Reconcile this copy of the EventListener and then write back any status
-	// updates
-	reconcileErr := c.reconcile(ctx, el)
-	if !equality.Semantic.DeepEqual(original.Status, el.Status) {
-		_, err = c.TriggersClientSet.TriggersV1alpha1().EventListeners(namespace).UpdateStatus(el)
-		if err != nil {
-			c.Logger.Warn("Failed to update EventListener status", err.Error())
-			return err
-		}
-	}
-	return reconcileErr
-}
+	logger := logging.FromContext(ctx)
 
-func (c *Reconciler) reconcile(ctx context.Context, el *v1alpha1.EventListener) error {
 	// We may be reading a version of the object that was stored at an older version
 	// and may not have had all of the assumed default specified.
 	el.SetDefaults(v1alpha1.WithUpgradeViaDefaulting(ctx))
 
-	// TODO(dibyom): Once #70 is merged, we should validate triggerTemplate here
-	// and update the StatusCondition
+	serviceReconcileError := r.reconcileService(logger, el)
+	deploymentReconcileError := r.reconcileDeployment(logger, el)
 
-	// TODO(vtereso): Create the resources within the reconciler, but restrict
-	// updates within an admission webhook instead. The reconciler is resolving
-	// behavior after it has been approved, which is from the wrong point of the
-	// lifecycle and presents inherent problems.
-	serviceReconcileError := c.reconcileService(el)
-	deploymentReconcileError := c.reconcileDeployment(el)
 	return wrapError(serviceReconcileError, deploymentReconcileError)
 }
 
-func reconcileObjectMeta(oldMeta *metav1.ObjectMeta, newMeta metav1.ObjectMeta) (updated bool) {
-	if !reflect.DeepEqual(oldMeta.Labels, newMeta.Labels) {
-		updated = true
-		oldMeta.Labels = newMeta.Labels
+// FinalizeKind cleans up associated logging config maps when an EventListener is deleted
+func (r *Reconciler) FinalizeKind(ctx context.Context, el *v1alpha1.EventListener) pkgreconciler.Event {
+	logger := logging.FromContext(ctx)
+	cfgs, err := r.eventListenerLister.EventListeners(el.Namespace).List(labels.Everything())
+	if err != nil {
+		return err
 	}
-	if !reflect.DeepEqual(oldMeta.OwnerReferences, newMeta.OwnerReferences) {
+	if len(cfgs) != 1 {
+		logger.Infof("Not deleting logging config map since more than one EventListener present in the namespace %s", el.Namespace)
+		return nil
+	}
+	// only one EL left
+	lastEL := cfgs[0]
+	if lastEL.Namespace == r.systemNamespace {
+		logger.Infof("Not deleting logging config map since EventListener is in the same namespace (%s) as the controller", el.Namespace)
+		return nil
+	}
+	if err = r.KubeClientSet.CoreV1().ConfigMaps(el.Namespace).Delete(eventListenerConfigMapName, &metav1.DeleteOptions{}); err != nil {
+		if errors.IsNotFound(err) {
+			return nil
+		}
+		return err
+	}
+	logger.Infof("Deleted logging config map since last EventListener in the namespace %s was deleted", lastEL.Namespace)
+	return nil
+}
+
+func reconcileObjectMeta(existing *metav1.ObjectMeta, desired metav1.ObjectMeta) (updated bool) {
+	if !reflect.DeepEqual(existing.Labels, desired.Labels) {
 		updated = true
-		oldMeta.OwnerReferences = newMeta.OwnerReferences
+		existing.Labels = desired.Labels
+	}
+
+	// TODO(dibyom): We should exclude propagation of some annotations such as `kubernetes.io/last-applied-revision`
+	if !reflect.DeepEqual(existing.Annotations, mergeMaps(existing.Annotations, desired.Annotations)) {
+		updated = true
+		existing.Annotations = desired.Annotations
+	}
+
+	if !reflect.DeepEqual(existing.OwnerReferences, desired.OwnerReferences) {
+		updated = true
+		existing.OwnerReferences = desired.OwnerReferences
 	}
 	return
 }
 
-func (c *Reconciler) reconcileService(el *v1alpha1.EventListener) error {
+func (r *Reconciler) reconcileService(logger *zap.SugaredLogger, el *v1alpha1.EventListener) error {
 	service := &corev1.Service{
 		ObjectMeta: generateObjectMeta(el),
 		Spec: corev1.ServiceSpec{
 			Selector: GenerateResourceLabels(el.Name),
 			Type:     el.Spec.ServiceType,
-			Ports: []corev1.ServicePort{
-				{
-					Name:     eventListenerServicePortName,
-					Protocol: corev1.ProtocolTCP,
-					Port:     int32(*ElPort),
-					TargetPort: intstr.IntOrString{
-						IntVal: int32(*ElPort),
-					},
-				},
+			Ports: []corev1.ServicePort{{
+				Name:     eventListenerServicePortName,
+				Protocol: corev1.ProtocolTCP,
+				Port:     int32(*ElPort),
+				TargetPort: intstr.IntOrString{
+					IntVal: int32(*ElPort),
+				}},
 			},
 		},
 	}
-	existingService, err := c.KubeClientSet.CoreV1().Services(el.Namespace).Get(el.Status.Configuration.GeneratedResourceName, metav1.GetOptions{})
+	existingService, err := r.serviceLister.Services(el.Namespace).Get(el.Status.Configuration.GeneratedResourceName)
 	switch {
 	case err == nil:
 		// Determine if reconciliation has to occur
 		updated := reconcileObjectMeta(&existingService.ObjectMeta, service.ObjectMeta)
+		el.Status.SetExistsCondition(v1alpha1.ServiceExists, nil)
+		el.Status.SetAddress(listenerHostname(service.Name, el.Namespace, *ElPort))
 		if !reflect.DeepEqual(existingService.Spec.Selector, service.Spec.Selector) {
 			existingService.Spec.Selector = service.Spec.Selector
 			updated = true
@@ -210,55 +211,55 @@ func (c *Reconciler) reconcileService(el *v1alpha1.EventListener) error {
 			updated = true
 		}
 		if updated {
-			if _, err := c.KubeClientSet.CoreV1().Services(el.Namespace).Update(existingService); err != nil {
-				c.Logger.Errorf("Error updating EventListener Service: %s", err)
+			if _, err := r.KubeClientSet.CoreV1().Services(el.Namespace).Update(existingService); err != nil {
+				logger.Errorf("Error updating EventListener Service: %s", err)
 				return err
 			}
-			c.Logger.Infof("Updated EventListener Service %s in Namespace %s", existingService.Namespace, el.Namespace)
+			logger.Infof("Updated EventListener Service %s in Namespace %s", existingService.Namespace, el.Namespace)
 		}
 	case errors.IsNotFound(err):
 		// Create the EventListener Service
-		_, err = c.KubeClientSet.CoreV1().Services(el.Namespace).Create(service)
+		_, err = r.KubeClientSet.CoreV1().Services(el.Namespace).Create(service)
 		el.Status.SetExistsCondition(v1alpha1.ServiceExists, err)
 		if err != nil {
-			c.Logger.Errorf("Error creating EventListener Service: %s", err)
+			logger.Errorf("Error creating EventListener Service: %s", err)
 			return err
 		}
 		el.Status.SetAddress(listenerHostname(service.Name, el.Namespace, *ElPort))
-		c.Logger.Infof("Created EventListener Service %s in Namespace %s", service.Name, el.Namespace)
+		logger.Infof("Created EventListener Service %s in Namespace %s", service.Name, el.Namespace)
 	default:
-		c.Logger.Error(err)
+		logger.Error(err)
 		return err
 	}
 	return nil
 }
 
-func (c *Reconciler) reconcileLoggingConfig(el *v1alpha1.EventListener) error {
-	_, err := c.KubeClientSet.CoreV1().ConfigMaps(el.Namespace).Get(eventListenerConfigMapName, metav1.GetOptions{})
-	if errors.IsNotFound(err) {
+func (r *Reconciler) reconcileLoggingConfig(logger *zap.SugaredLogger, el *v1alpha1.EventListener) error {
+	if _, err := r.configmapLister.ConfigMaps(el.Namespace).Get(eventListenerConfigMapName); errors.IsNotFound(err) {
 		// create default config-logging ConfigMap
-		_, err = c.KubeClientSet.CoreV1().ConfigMaps(el.Namespace).Create(defaultLoggingConfigMap())
-		if err != nil {
-			c.Logger.Errorf("Failed to create logging config: %s.  EventListener won't start.", err)
+		if _, err := r.KubeClientSet.CoreV1().ConfigMaps(el.Namespace).Create(defaultLoggingConfigMap()); err != nil {
+			logger.Errorf("Failed to create logging config: %s.  EventListener won't start.", err)
 			return err
 		}
 	} else if err != nil {
-		c.Logger.Errorf("Error retrieving ConfigMap %q: %s", eventListenerConfigMapName, err)
+		logger.Errorf("Error retrieving ConfigMap %q: %s", eventListenerConfigMapName, err)
 		return err
 	}
 	return nil
 }
 
-func (c *Reconciler) reconcileDeployment(el *v1alpha1.EventListener) error {
+func (r *Reconciler) reconcileDeployment(logger *zap.SugaredLogger, el *v1alpha1.EventListener) error {
 	// check logging config, create if it doesn't exist
-	err := c.reconcileLoggingConfig(el)
-	if err != nil {
-		c.Logger.Error(err)
+	if err := r.reconcileLoggingConfig(logger, el); err != nil {
+		logger.Error(err)
 		return err
 	}
 
-	labels := mergeLabels(el.Labels, GenerateResourceLabels(el.Name))
-	var replicas int32 = 1
+	labels := mergeMaps(el.Labels, GenerateResourceLabels(el.Name))
+	var replicas = ptr.Int32(1)
+	if el.Spec.Replicas != nil {
+		replicas = el.Spec.Replicas
+	}
 	container := corev1.Container{
 		Name:  "event-listener",
 		Image: *elImage,
@@ -277,6 +278,17 @@ func (c *Reconciler) reconcileDeployment(el *v1alpha1.EventListener) error {
 			PeriodSeconds:    int32(*PeriodSeconds),
 			FailureThreshold: int32(*FailureThreshold),
 		},
+		ReadinessProbe: &corev1.Probe{
+			Handler: corev1.Handler{
+				HTTPGet: &corev1.HTTPGetAction{
+					Path:   "/live",
+					Scheme: corev1.URISchemeHTTP,
+					Port:   intstr.FromInt((*ElPort)),
+				},
+			},
+			PeriodSeconds:    int32(*PeriodSeconds),
+			FailureThreshold: int32(*FailureThreshold),
+		},
 		Args: []string{
 			"-el-name", el.Name,
 			"-el-namespace", el.Namespace,
@@ -298,7 +310,7 @@ func (c *Reconciler) reconcileDeployment(el *v1alpha1.EventListener) error {
 	deployment := &appsv1.Deployment{
 		ObjectMeta: generateObjectMeta(el),
 		Spec: appsv1.DeploymentSpec{
-			Replicas: &replicas,
+			Replicas: replicas,
 			Selector: &metav1.LabelSelector{
 				MatchLabels: GenerateResourceLabels(el.Name),
 			},
@@ -307,6 +319,8 @@ func (c *Reconciler) reconcileDeployment(el *v1alpha1.EventListener) error {
 					Labels: labels,
 				},
 				Spec: corev1.PodSpec{
+					Tolerations:        el.Spec.PodTemplate.Tolerations,
+					NodeSelector:       el.Spec.PodTemplate.NodeSelector,
 					ServiceAccountName: el.Spec.ServiceAccountName,
 					Containers:         []corev1.Container{container},
 
@@ -324,15 +338,20 @@ func (c *Reconciler) reconcileDeployment(el *v1alpha1.EventListener) error {
 			},
 		},
 	}
-	existingDeployment, err := c.KubeClientSet.AppsV1().Deployments(el.Namespace).Get(el.Status.Configuration.GeneratedResourceName, metav1.GetOptions{})
+	existingDeployment, err := r.deploymentLister.Deployments(el.Namespace).Get(el.Status.Configuration.GeneratedResourceName)
 	switch {
 	case err == nil:
 		el.Status.SetDeploymentConditions(existingDeployment.Status.Conditions)
+		el.Status.SetExistsCondition(v1alpha1.DeploymentExists, nil)
+
 		// Determine if reconciliation has to occur
 		updated := reconcileObjectMeta(&existingDeployment.ObjectMeta, deployment.ObjectMeta)
-		if existingDeployment.Spec.Replicas == nil || *existingDeployment.Spec.Replicas == 0 {
-			existingDeployment.Spec.Replicas = &replicas
-			updated = true
+		if *existingDeployment.Spec.Replicas != *deployment.Spec.Replicas {
+			if el.Spec.Replicas != nil {
+				existingDeployment.Spec.Replicas = replicas
+				updated = true
+			}
+			// if no replicas found as part of el.Spec then replicas from existingDeployment will be considered
 		}
 		if existingDeployment.Spec.Selector != deployment.Spec.Selector {
 			existingDeployment.Spec.Selector = deployment.Spec.Selector
@@ -346,6 +365,14 @@ func (c *Reconciler) reconcileDeployment(el *v1alpha1.EventListener) error {
 			existingDeployment.Spec.Template.Spec.ServiceAccountName = deployment.Spec.Template.Spec.ServiceAccountName
 			updated = true
 		}
+		if !reflect.DeepEqual(existingDeployment.Spec.Template.Spec.Tolerations, deployment.Spec.Template.Spec.Tolerations) {
+			existingDeployment.Spec.Template.Spec.Tolerations = deployment.Spec.Template.Spec.Tolerations
+			updated = true
+		}
+		if !reflect.DeepEqual(existingDeployment.Spec.Template.Spec.NodeSelector, deployment.Spec.Template.Spec.NodeSelector) {
+			existingDeployment.Spec.Template.Spec.NodeSelector = deployment.Spec.Template.Spec.NodeSelector
+			updated = true
+		}
 		if len(existingDeployment.Spec.Template.Spec.Containers) == 0 ||
 			len(existingDeployment.Spec.Template.Spec.Containers) > 1 {
 			existingDeployment.Spec.Template.Spec.Containers = []corev1.Container{container}
@@ -381,24 +408,24 @@ func (c *Reconciler) reconcileDeployment(el *v1alpha1.EventListener) error {
 			}
 		}
 		if updated {
-			if _, err := c.KubeClientSet.AppsV1().Deployments(el.Namespace).Update(existingDeployment); err != nil {
-				c.Logger.Errorf("Error updating EventListener Deployment: %s", err)
+			if _, err := r.KubeClientSet.AppsV1().Deployments(el.Namespace).Update(existingDeployment); err != nil {
+				logger.Errorf("Error updating EventListener Deployment: %s", err)
 				return err
 			}
-			c.Logger.Infof("Updated EventListener Deployment %s in Namespace %s", existingDeployment.Name, el.Namespace)
+			logger.Infof("Updated EventListener Deployment %s in Namespace %s", existingDeployment.Name, el.Namespace)
 		}
 	case errors.IsNotFound(err):
 		// Create the EventListener Deployment
-		deployment, err = c.KubeClientSet.AppsV1().Deployments(el.Namespace).Create(deployment)
+		deployment, err = r.KubeClientSet.AppsV1().Deployments(el.Namespace).Create(deployment)
 		el.Status.SetExistsCondition(v1alpha1.DeploymentExists, err)
 		if err != nil {
-			c.Logger.Errorf("Error creating EventListener Deployment: %s", err)
+			logger.Errorf("Error creating EventListener Deployment: %s", err)
 			return err
 		}
 		el.Status.SetDeploymentConditions(deployment.Status.Conditions)
-		c.Logger.Infof("Created EventListener Deployment %s in Namespace %s", deployment.Name, el.Namespace)
+		logger.Infof("Created EventListener Deployment %s in Namespace %s", deployment.Name, el.Namespace)
 	default:
-		c.Logger.Error(err)
+		logger.Error(err)
 		return err
 	}
 	return nil
@@ -421,13 +448,14 @@ func generateObjectMeta(el *v1alpha1.EventListener) metav1.ObjectMeta {
 		Namespace:       el.Namespace,
 		Name:            el.Status.Configuration.GeneratedResourceName,
 		OwnerReferences: []metav1.OwnerReference{*el.GetOwnerReference()},
-		Labels:          mergeLabels(el.Labels, GenerateResourceLabels(el.Name)),
+		Labels:          mergeMaps(el.Labels, GenerateResourceLabels(el.Name)),
+		Annotations:     el.Annotations,
 	}
 }
 
-// mergeLabels merges the values in the passed maps into a new map.
+// mergeMaps merges the values in the passed maps into a new map.
 // Values within m2 potentially clobber m1 values.
-func mergeLabels(m1, m2 map[string]string) map[string]string {
+func mergeMaps(m1, m2 map[string]string) map[string]string {
 	merged := make(map[string]string, len(m1)+len(m2))
 	for k, v := range m1 {
 		merged[k] = v
